"""
HullRad
Version 7 (Python 3 edition)
Calculates hydrodynamic coefficients for a biomolecular structure.
AUTHOR: Patrick Fleming

Release Notes:
Version 5 (The program as described in Fleming and Fleming, BJ, 2018)
    Protein
    Nucleic Acids
Version 6 (2019)
    Includes many saccharides for glycosylation
    Rg is calculated from all atom model, not reduced model as in version 5
    Includes option to use numpy and scipy
    Asphericity is calculated from gyration tensor (requires numpy)
    Option to use either qconvex from qhull or ConvexHull from scipy
        (Many thanks to Chad Brautigam for implementation)
Version 7 (2020)
    Ported to Python 3 (See HullRadV7_3.py, this edition is Python 2)
    Addition of chain ID
    Includes several detergents for analysis of protein/detergent micelles.
        (See DT_data below for list.)


If you publish work that uses this code please cite:
Fleming, PJ and Fleming, KG "HullRad: Fast Calculations of Folded and Disordered
Protein and Nucleic Acid Hydrogynamic Properties", Biophysical Journal, 2018,
114:856-869
DOI: 10.1016/j.bpj.2018.01.002


The author would like to acknowledge the work of Jose Garcia de la Torre who led
the way in development of algorithms for calculating macromolecular
hydrodynamic properties.

--------------------------------------------------------------------------------
Output looks like:

 pdb8rat.ent
 HYDROLASE (NUCLEIC ACID,RNA)            13-AUG-91   8RAT
  #Amino Acids    :          124
  #Nucleotides    :            0
  #Saccharides    :            0
  #Detergents     :            0
  M               :        13692     g/mol
  v_bar           :        0.710     mL/g
  R(Anhydrous)    :        15.68     Angstroms
  Axial Ratio     :         1.31
  f/fo            :         1.21
  Dt              :       1.13e-06   cm^2/s
  R(Translation)  :        18.99     Angstroms
  s               :       1.84e-13   sec
  [eta]           :         3.16     cm^3/g
  Dr              :       1.91e+07   s^-1
  R(Rotation)     :        20.36     Angstroms
  tauC            :         8.84     ns (from R_rotation)

To write out the unified side chain pseudo-atom model uncomment the two lines
near the bottom labeled, # Write out model (~line 529)

--------------------------------------------------------------------------------
Requires python version2.7

Uses PDB file format as input.

If your python2.7 installation includes numpy and scipy YOU DON'T NEED ANYTHING ELSE.
This is the preferred way to go.

If you don't have numpy and scipy installed, you will need a separate program to 
    calculate the convex hull - called qconvex. 
This is a program in the qhull suite of programs.

Qhull may be downloaded from http://www.qhull.org/download.
    For UNIX download "Qhull_2015.2 for Unix"
    tar xzvf qhull-2015-src-7.2.0.tgz
    cd qhull-2015.2/
    make
    sudo make install
    and the path for qconvex will be /usr/local/bin/qconvex

OS X qhull binaries are available from MacPorts package manager. 
    type "sudo port install qhull".
    and the path for qconvex will be /opt/local/bin/qconvex

OS X binaries are also available from Fink package manager.

The default in this script is the UNIX path (/usr/local/bin/qconvex) but you may 
have to change it if you are using OS X.

After installing qhull go to: ### Edit the path to qconvex ### below (~line 476) and
change /usr/local/bin/qconvex to /opt/local/bin/qconvex if that is where qconvex is.

If you are using OS X but are familiar with UNIX and have gcc installed, the UNIX
instructions above work fine on a Mac.

FOR WINDOWS USERS:
There are precompiled versions of the qhull programs available for Windows 7 and up
at http://www.qhull.org/download
Unpack qhull in a directory of choice and ### Edit the path to qconvex ### below (~line 330)
to include the path where your executable was unpacked. 

--------------------------------------------------------------------------------
This is free and unencumbered software released into the public domain.

Anyone is free to copy, modify, publish, use, compile, sell, or
distribute this software, either in source code form or as a compiled
binary, for any purpose, commercial or non-commercial, and by any
means.

In jurisdictions that recognize copyright laws, the author or authors
of this software dedicate any and all copyright interest in the
software to the public domain. We make this dedication for the benefit
of the public at large and to the detriment of our heirs and
successors. We intend this dedication to be an overt act of
relinquishment in perpetuity of all present and future rights to this
software under copyright law.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

For more information, please refer to <http://unlicense.org>

RB 2021
CHANGES of the original version 7.
- scipy, numpy required, removed according 2.7 code
- removed python3 conversion code to be only python3 compatible (python2.7 is not maintained anymore)
- changes to use it in jscatter.bio
- removed some not needed (confusing) comments above

"""

import math
import os
import string
import sys
import io

import numpy as np
from scipy.spatial import ConvexHull
import MDAnalysis

# Hack for 2to3 conversion
for attr in ['capitalize', 'count', 'find', 'index', 'lower', 'replace', 'rstrip', 'split', 'strip', 'upper', 'zfill']:
    setattr(string, attr, getattr(str, attr))
string.letters = string.ascii_letters
string.lowercase = string.ascii_lowercase
string.join = lambda words, sep=' ': sep.join(words)
string.atoi = int

# Amino Acid mass, volumes,partial specific volumes and unified side chain sphere radius
# Mass is minus water (from SEDNTERP database)
# Volumes from Cohn and Edsall (1943) as corrected by Perkins_EJB_1986 
# vbar is 0.60224(V/M)
#          Mass        Volume       vbar        Radius of equiv. sphere of side chain volume
AA_data = {
    'ALA': (71.0939, 87.2, 0.74, 1.852),
    'ARG': (156.203, 188.2, 0.73, 3.123),
    'ASN': (114.119, 120.1, 0.63, 2.422),
    'ASP': (115.104, 115.4, 0.60, 2.356),
    'CYS': (103.16, 106.7, 0.62, 2.224),
    'GLN': (128.16, 145.1, 0.68, 2.722),
    'GLU': (129.131, 140.9, 0.66, 2.676),
    'GLY': (57.0669, 60.6, 0.64, 0.000),
    'HIS': (137.157, 152.4, 0.67, 2.798),
    'HSE': (137.157, 152.4, 0.67, 2.798),  # CHARMM:neutral His, proton on NE2
    'HSP': (137.157, 152.4, 0.67, 2.798),  # CHARMM:Protonated His
    'HSD': (137.157, 152.4, 0.67, 2.798),  # CHARMM:neutral HIS, proton on ND1
    'ILE': (113.175, 168.9, 0.90, 2.957),
    'LEU': (113.175, 168.9, 0.90, 2.957),
    'LYS': (128.19, 174.3, 0.82, 3.005),
    'MET': (131.215, 163.1, 0.75, 2.903),
    'MSE': (178.125, 163.1, 0.75, 2.903),
    'PHE': (147.192, 187.9, 0.77, 3.121),
    'PRO': (97.132, 122.4, 0.76, 2.453),
    'SER': (87.093, 91.0, 0.63, 1.936),
    'THR': (101.12, 117.4, 0.70, 2.385),
    'TRP': (186.229, 228.5, 0.74, 3.422),
    'TYR': (163.192, 192.1, 0.71, 3.155),
    'VAL': (99.148, 141.4, 0.86, 2.682)
}

# From voss_jmb_2005.pdf and nadassy_nar_2001.pdf
# Data for nucleoside, = nucleotide - 79.95 for PO3H
# Phosphate for na lacking terminal phosphates
# vbar is 0.60224(V/M)
#          Mass         Volume   vbar      ND
NA_data = {
    '  A': (267.25, 272.3, 0.614, 0.00),  # Adenosine
    '1MA': (281.25, 299.3, 0.641, 0.00),  # 6-HYDRO-1-METHYLADENOSINE
    'MIA': (383.45, 385.1, 0.605, 0.00),  # 2-METHYLTHIO-N6-ISOPENTENYL-ADENOSINE
    '  C': (243.22, 248.1, 0.614, 0.00),  # Cytidine
    '5MC': (257.22, 275.1, 0.644, 0.00),  # 5-METHYLCYTIDINE
    'OMC': (257.22, 275.1, 0.644, 0.00),  # O2'-METHYLYCYTIDINE
    '  G': (283.24, 279.0, 0.593, 0.00),  # Guanosine
    '2MG': (297.27, 306.0, 0.620, 0.00),  # 2N-METHYLGUANOSINE
    '7MG': (299.27, 306.0, 0.620, 0.00),  # 7N-METHYL-8-HYDROGUANOSINE
    'M2G': (311.27, 333.0, 0.644, 0.00),  # N2-DIMETHYLGUANOSINE
    'OMG': (297.27, 306.0, 0.620, 0.00),  # O2'-METHYLGUANOSINE
    'YYG': (508.54, 500.0, 0.593, 0.00),  # MODIFIED GUANOSINE
    ' YG': (428.55, 427.1, 0.600, 0.00),  # WYBUTOSINE
    '  U': (244.20, 243.9, 0.602, 0.00),  # Uracil
    '5MU': (258.20, 270.9, 0.632, 0.00),  # 5-METHYLURIDINE
    '4SU': (260.35, 270.5, 0.626, 0.00),  # 4-THIOURIDINE
    'H2U': (246.20, 243.9, 0.602, 0.00),  # 5,6-DIHYDROURIDINE
    'PSU': (244.20, 243.9, 0.602, 0.00),  # PSEUDOURIDINE
    '  I': (268.23, 273.0, 0.613, 0.00),  # Inosine
    ' DA': (251.25, 271.0, 0.650, 0.00),  # 2'-DEOXYADENOSINE
    ' DC': (227.22, 246.8, 0.654, 0.00),  # 2'-DEOXYCYTIDINE
    ' DG': (267.25, 277.7, 0.626, 0.00),  # 2'-DEOXYGUANOSINE
    ' DT': (242.23, 264.4, 0.657, 0.00),  # 2'-DEOXYTHYMIDINE
    ' DI': (252.22, 271.7, 0.649, 0.00),  # 2'-DEOXYINOSINE
    'PO2': (62.97, 52.4, 0.501, 0.00)
}

# RB: a bug in MDAnalysis pdb files ignores th spaces in front, we include these here
for k in list(NA_data.keys()):
    if k[0] == ' ':
        NA_data.update({k[1:]+' ': NA_data[k]})
    elif k[:2] == '  ':
        NA_data.update({k[2]+'  ': NA_data[k]})

# For default saccharides use average vbar = 0.63
# Schuster, TM & Laue, TM, Eds, Modern Analytical UC, Springer, 2012
# or if found in, use
# Schuck, P. Zhao, H. Brautigam, C.A. and Ghirlando, R.
# Basic Principles of Analytical Ultracentrifugation
# CRC Press, 2016
#
# V = (vbar*M)/0.60224
#          Mass         Volume   vbar      ND
GL_data = {
    'NG6': (301.27, 315.2, 0.630, 0.00),  # N-ACETYL-D-GALACTOSAMINE 6-SULFATE
    'NAG': (221.21, 231.4, 0.630, 0.00),  # N-ACETYL-D-GLUCOSAMINE
    'BM3': (221.21, 231.4, 0.630, 0.00),  # 2-(ACETYLAMINO)-2-DEOXY-ALPHA-D-MANNOPYRANOSE
    'NGA': (221.21, 231.4, 0.684, 0.00),  # N-ACETYL-D-GALACTOSAMINE
    'GCU': (194.14, 203.1, 0.630, 0.00),  # D-GLUCURONIC ACID
    'IDR': (194.14, 203.1, 0.630, 0.00),  # L-IDURONIC ACID
    'BMA': (180.16, 188.4, 0.607, 0.00),  # BETA-D-MANNOSE
    'MAN': (180.16, 188.4, 0.607, 0.00),  # ALPHA-D-MANNOSE
    'GAL': (180.16, 188.4, 0.622, 0.00),  # BETA-D-GALACTOSE
    'GLA': (180.16, 188.4, 0.622, 0.00),  # ALPHA D-GALACTOSE
    'GLC': (180.16, 188.4, 0.622, 0.00),  # ALPHA-D-GLUCOSE
    'BGC': (180.16, 188.4, 0.622, 0.00),  # BETA-D-GLUCOSE
    'AOS': (180.16, 188.4, 0.630, 0.00),  # D-ALLOSE
    'GCS': (179.17, 187.4, 0.630, 0.00),  # Glucosamine
    'RAM': (164.16, 171.7, 0.630, 0.00),  # ALPHA-L-RHAMNOSE
    'FUC': (164.16, 171.7, 0.671, 0.00),  # ALPHA-L-FUCOSE
    'SIA': (309.27, 299.9, 0.584, 0.00)  # O-SIALIC ACID
}

# Detergents
# From Schuck, P. Zhao, H. Brautigam, C.A. and Ghirlando, R.
# Basic Principles of Analytical Ultracentrifugation
# CRC Press, 2016
#
# DOC from Durschlag, H.
# Specific Volumes of Biological Macromolecules and Some Other
#   Molecules of Biological Interest.
# in Thermodynamic Data for Biochemistry and Biotechnology, ed. H-J Hinz
# Springer-Verlag, 1986
#
# LMT from Suarez et al.
# JBC 259:13791, 1984
#
# FOS (DPC) from Kochendoerfer, et al
# Biochemistry 38:11905, 1999
#  But see Lauterwin, BBA 556:244, 1979 for vbar = 0.937
#
# V = (vbar*M)/0.60224
#          Mass         Volume   vbar      ND
DT_data = {
    'SB3': (335.50, 533.9, 0.957, 0.00),  # n-Dodecyl-N,N-dimethyl-3-ammonio-1-propanesulfonate
    'LMT': (510.62, 691.0, 0.820, 0.00),  # DODECYL-BETA-D-MALTOSIDE
    'BOG': (417.02, 594.8, 0.859, 0.00),  # B-OCTYLGLUCOSIDE
    'LDA': (229.40, 429.7, 1.128, 0.00),  # LAURYL DIMETHYLAMINE-N-OXIDE
    'SDS': (266.40, 384.8, 0.880, 0.00),  # DODECYL SULFATE
    'DXC': (392.47, 507.0, 0.778, 0.00),  # DEOXYCHOLATE
    'FOS': (351.46, 548.6, 0.940, 0.00)  # DODECYLPHOSPHOCHOLINE
}


def distance(ax, ay, az, bx, by, bz):
    # Euclidean distance between two atoms
    return math.sqrt((ax - bx) ** 2.0 + (ay - by) ** 2.0 + (az - bz) ** 2.0)


def get_coords(model_array):
    # Get x,y,z coords of CA and CB atoms
    coords = []
    for row in range(len(model_array)):
        coords.append((float(model_array[row][5]), float(model_array[row][6]), float(model_array[row][7])))
    return np.array(coords)


def model_from_pdb(file):
    # Makes a reduced atom  model of the pdb file
    # This is the model used to make the convex hull
    # For proteins the CB is displaced along the CA-CB vector a distance 
    #   equal to the radius of a sphere equal to the volume of the side chain
    # For nucleic acids only the phosphate, oxygen and nitrogen atoms are used 
    # For saccharides only the phosphate, oxygen and nitrogen atoms are used
    # For detergents only the oxygen and nitrogen atoms are used

    # Read the PDB file
    if isinstance(file, str):
        with open(file, 'r') as f:
            data = f.readlines()
    else:
        # list of lines from pdb readlines
        data = file

    if not data:
        raise TypeError('No data read from file.')

    # Initialize all relevant atom record array
    # Used for Rg
    all_atm_rec = []
    # Initial protein atoms from PDB file
    prot_rec = []
    # Initial nucleic acid atoms from PDB file
    na_rec = []
    # Initial saccharide atoms from PDB file
    gl_rec = []
    # Initial detergent atoms from PDB file
    dt_rec = []
    # Reduced list of atoms
    atom_rec = []
    struc_name = ' '
    num_MG = 0
    num_MN = 0

    # Get all relevant atoms even if in wrong order
    for line in data:
        # Get name of structure
        if line[:6] == 'HEADER':
            struc_name = line.strip()
            print(struc_name[9:])

        elif line[:4] == 'ATOM':
            resname = line[17:20]
            if resname != 'TIP':
                all_atm_rec.append(line)
                if resname in AA_data and (line[13:15] == 'N ' or
                                           line[13:15] == 'CA' or
                                           line[13:15] == 'C ' or
                                           line[13:15] == 'O ' or
                                           line[13:16] == 'OT1' or
                                           line[13:15] == 'CB') and \
                        (line[16] != 'B') and \
                        (line[:3] != 'END'):
                    prot_rec.append(line)
                # Nucleic acids
                elif resname in NA_data and (line[13:14] == 'N' or
                                             line[13:14] == 'O' or line[13:14] == 'P'):
                    na_rec.append(line)
                # Saccharides
                elif resname in GL_data and (line[13:14] == 'N' or
                                             line[13:14] == 'O' or line[13:14] == 'P'):
                    gl_rec.append(line)
                # Detergents
                elif resname in DT_data and (line[13:14] == 'N' or
                                             line[13:14] == 'O' or line[14:15] == 'O'):
                    dt_rec.append(line)

        elif line[:6] == 'HETATM':
            resname = line[17:20]
            if resname != 'TIP':
                for rec in NA_data.items():
                    if resname == rec[0]:
                        all_atm_rec.append(line)
                for rec in GL_data.items():
                    if resname == rec[0]:
                        all_atm_rec.append(line)
                for rec in DT_data.items():
                    if resname == rec[0]:
                        all_atm_rec.append(line)
                if line[12:14] == 'MG':
                    num_MG += 1
                elif line[12:14] == 'MN':
                    num_MN += 1
                # Some nucleic acids are in HETATM (!)
                elif resname in NA_data and (line[13:14] == 'N' or
                                             line[13:14] == 'O' or line[13:14] == 'P'):
                    na_rec.append(line)
                # Some glycosylations are in HETATM (!)
                elif resname in GL_data and (line[13:14] == 'N' or
                                             line[13:14] == 'O' or line[13:14] == 'P'):
                    gl_rec.append(line)
                # Some detergents are in HETATM (!)
                elif resname in DT_data and (line[13:14] == 'N' or
                                             line[13:14] == 'O'):
                    dt_rec.append(line)

    # Get only the CB atoms from the protein
    cb_rec = []
    for line in prot_rec:
        if (line[13:15] == 'CB') and \
                (line[0:3] != 'END'):
            cb_rec.append(line)

    # Count the CB atoms and add to atom list in correct order
    num_cb = len(cb_rec)
    line_counter = 0
    cb_counter = 0
    for line in prot_rec:
        if (line[13:15] == 'N ' or
                line[13:15] == 'CA' or
                line[13:15] == 'C ' or
                line[13:15] == 'O ' or
                line[13:16] == 'OT1'):
            atom_rec.append(line)
            line_counter += 1
            if (math.fmod(line_counter, 4) == 0 and line[17:20] != 'GLY'
                    and cb_counter < num_cb):
                atom_rec.append((cb_rec[cb_counter]))
                cb_counter += 1
            elif (math.fmod(line_counter, 4) == 0 and line[17:20] != 'GLY'
                  and cb_counter >= num_cb):
                print(' Found non-gly residue with no CB atom')
                sys.exit()

    # Put atom info into model array
    num_atoms = len(atom_rec)
    model_array = [[' X' for j in range(8)] for i in range(num_atoms)]
    for row in range(len(model_array)):
        model_array[row][0] = row
        model_array[row][1] = (atom_rec[row][11:16])
        model_array[row][2] = (atom_rec[row][17:20])
        if atom_rec[row][20:22] == '  ':
            model_array[row][3] = ' X'
        else:
            model_array[row][3] = (atom_rec[row][20:22])
        model_array[row][4] = (atom_rec[row][22:26])
        model_array[row][5] = (atom_rec[row][30:38])
        model_array[row][6] = (atom_rec[row][38:46])
        model_array[row][7] = (atom_rec[row][46:54])

    #   for line in range(len(model_array)):
    #       print model_array[line]

    # Make unified side chain as CB
    for row in range(len(model_array)):
        if model_array[row][1] == '  CA ' and model_array[row][2] != 'GLY':
            rnam = model_array[row][2]
            # CA coords
            ca_x = float(model_array[row][5])
            ca_y = float(model_array[row][6])
            ca_z = float(model_array[row][7])
            # CB coords (third next coord: C,O,CB)
            cb_x = float(model_array[row + 3][5])
            cb_y = float(model_array[row + 3][6])
            cb_z = float(model_array[row + 3][7])
            # CA-CB distance
            dx = cb_x - ca_x
            dy = cb_y - ca_y
            dz = cb_z - ca_z
            ca_cb_dist = distance(ca_x, ca_y, ca_z, cb_x, cb_y, cb_z)
            # Extend CB outward along CA-CB vector
            aa_mass, vol_aa, vbar_aa, sc_rad_aa = AA_data[rnam]
            new_cb_x = ca_x + (sc_rad_aa * dx / ca_cb_dist)
            new_cb_y = ca_y + (sc_rad_aa * dy / ca_cb_dist)
            new_cb_z = ca_z + (sc_rad_aa * dz / ca_cb_dist)
            model_array[row + 3][5] = str(new_cb_x)
            model_array[row + 3][6] = str(new_cb_y)
            model_array[row + 3][7] = str(new_cb_z)

    # Put nucleic acid atoms into initial array
    num_atoms = len(na_rec)
    na_array = [['X' for j in range(8)] for i in range(num_atoms)]
    for row in range(len(na_array)):
        na_array[row][0] = row
        na_array[row][1] = (na_rec[row][11:16])
        na_array[row][2] = (na_rec[row][17:20])
        na_array[row][3] = (na_rec[row][20:22])
        na_array[row][4] = (na_rec[row][22:26])
        na_array[row][5] = (na_rec[row][30:38])
        na_array[row][6] = (na_rec[row][38:46])
        na_array[row][7] = (na_rec[row][46:54])

    # Add nucleic acid atoms to model array
    for row in range(len(na_array)):
        model_array.append(na_array[row])

    # Put saccharide atoms into initial array
    num_atoms = len(gl_rec)
    gl_array = [['X' for j in range(8)] for i in range(num_atoms)]
    for row in range(len(gl_array)):
        gl_array[row][0] = row
        gl_array[row][1] = (gl_rec[row][11:16])
        gl_array[row][2] = (gl_rec[row][17:20])
        gl_array[row][3] = (gl_rec[row][20:22])
        gl_array[row][4] = (gl_rec[row][22:26])
        gl_array[row][5] = (gl_rec[row][30:38])
        gl_array[row][6] = (gl_rec[row][38:46])
        gl_array[row][7] = (gl_rec[row][46:54])

    # Add saccharide atoms into model array
    for row in range(len(gl_array)):
        model_array.append(gl_array[row])

    # Put detergent atoms into initial array
    num_atoms = len(dt_rec)
    dt_array = [['X' for j in range(8)] for i in range(num_atoms)]
    for row in range(len(dt_array)):
        dt_array[row][0] = row
        dt_array[row][1] = (dt_rec[row][11:16])
        dt_array[row][2] = (dt_rec[row][17:20])
        dt_array[row][3] = (dt_rec[row][20:22])
        dt_array[row][4] = (dt_rec[row][22:26])
        dt_array[row][5] = (dt_rec[row][30:38])
        dt_array[row][6] = (dt_rec[row][38:46])
        dt_array[row][7] = (dt_rec[row][46:54])

    # Add detergent atoms into model array
    for row in range(len(dt_array)):
        model_array.append(dt_array[row])

    return all_atm_rec, num_MG, num_MN, model_array


def write_pdb(model_array, filename):
    # Write out reduced atom model in PDB format for display

    # pdb format string
    pdbfmt = 'ATOM  %5d%3s %3s%2s%4d    %8.3f%8.3f%8.3f  0.00  0.00\n'

    with open(filename, 'w') as ff:
        for i in range(len(model_array)):
            a = model_array[i]
            ff.write(pdbfmt % (int(a[0]), a[1], a[2], a[3], int(a[4]), float(a[5]), float(a[6]), float(a[7])))
        write('END\n')


def Sved(all_atm_rec, num_MG, num_MN, model_array):
    #
    # Main function: Does most things and calls HullVolume
    #
    # Use Cohn-Edsall eq. to calc protein specific volume
    # vbar_prot = sum(ni*Mi*vbar_aa) / (ni*Mi) where n is number, M is mass
    #
    prot_mol_mass = 0.0
    numerator = 0.0
    AA = 0
    NA = 0
    GL = 0
    DT = 0

    # Radius of Gyration
    # Calc center of mass
    X = 0.0
    Y = 0.0
    Z = 0.0
    tot = 0.0
    for row in range(len(all_atm_rec)):
        X = X + (float(all_atm_rec[row][30:38]))
        Y = Y + (float(all_atm_rec[row][38:46]))
        Z = Z + (float(all_atm_rec[row][46:54]))
        tot += 1

    com_x = (X / tot)
    com_y = (Y / tot)
    com_z = (Z / tot)

    Rg2 = 0.0
    for row in range(len(all_atm_rec)):
        Rg2 += ((distance(com_x, com_y, com_z, float(all_atm_rec[row][30:38]),
                          float(all_atm_rec[row][38:46]), float(all_atm_rec[row][46:54]))) ** 2)
    Rg = math.sqrt(Rg2 / tot)

    # Asphericity
    # 0 for sphere, 1 for rod
    Ixx, Ixy, Ixz, Iyy, Iyz, Izz = 0, 0, 0, 0, 0, 0
    for row in range(len(all_atm_rec)):
        Ixx += ((float(all_atm_rec[row][30:38])) - com_x) * ((float(all_atm_rec[row][30:38])) - com_x)
        Ixy += ((float(all_atm_rec[row][30:38])) - com_x) * ((float(all_atm_rec[row][38:46])) - com_y)
        Ixz += ((float(all_atm_rec[row][30:38])) - com_x) * ((float(all_atm_rec[row][46:54])) - com_z)
        Iyy += ((float(all_atm_rec[row][38:46])) - com_y) * ((float(all_atm_rec[row][38:46])) - com_y)
        Iyz += ((float(all_atm_rec[row][38:46])) - com_y) * ((float(all_atm_rec[row][46:54])) - com_z)
        Izz += ((float(all_atm_rec[row][46:54])) - com_z) * ((float(all_atm_rec[row][46:54])) - com_z)

    Ixx = Ixx / row
    Iyy = Iyy / row
    Izz = Izz / row
    Ixy = Ixy / row
    Ixz = Ixz / row
    Iyz = Iyz / row

    gyration_tensor = [[Ixx, Ixy, Ixz], [Ixy, Iyy, Iyz], [Ixz, Iyz, Izz]]
    evals, evecs = np.linalg.eig(gyration_tensor)
    L1 = evals[0]
    L2 = evals[1]
    L3 = evals[2]

    asphr = ((L1 - L2) ** 2 + (L2 - L3) ** 2 + (L1 - L3) ** 2) / (2.0 * ((L1 + L2 + L3) ** 2))

    # Calc masses for all
    last_rnum = 99999
    for row in range(len(model_array)):
        if model_array[row][1] == '  CA ':
            rnam = model_array[row][2]
            aa_mass, vol_aa, vbar_aa, sc_rad_aa = AA_data[rnam]
            numerator += (aa_mass * vbar_aa)
            prot_mol_mass += aa_mass
            AA = AA + 1

        # DNA, RNA, Saccharides, Detergents
        else:
            rnam = model_array[row][2]
            chid = model_array[row][3]
            rnum = model_array[row][4]
            if rnum != last_rnum:
                for rec in NA_data.items():
                    if rnam == rec[0]:
                        aa_mass, vol_aa, vbar_aa, sc_rad_aa = NA_data[rnam]
                        numerator += (aa_mass * vbar_aa)
                        prot_mol_mass += aa_mass
                        last_rnum = rnum
                        NA = NA + 1
                        # Check for nucleotide instead of nucleoside
                        if model_array[row][1] == "  P  ":
                            numerator += (62.97 * 0.501)
                            prot_mol_mass += 62.97
                for rec in GL_data.items():
                    if rnam == rec[0]:
                        aa_mass, vol_aa, vbar_aa, sc_rad_aa = GL_data[rnam]
                        numerator += (aa_mass * vbar_aa)
                        prot_mol_mass += aa_mass
                        last_rnum = rnum
                        GL = GL + 1
                for rec in DT_data.items():
                    if rnam == rec[0]:
                        aa_mass, vol_aa, vbar_aa, sc_rad_aa = DT_data[rnam]
                        numerator += (aa_mass * vbar_aa)
                        prot_mol_mass += aa_mass
                        last_rnum = rnum
                        DT = DT + 1

    # Add weight of water for N-term and C-term of protein
    if AA > 0:
        prot_mol_mass += 18.0
    # If nucleic acid add weight of MG and MN
    if NA > 0:
        ion_mass = (num_MG * 24.305) + (num_MN * 54.938)
        prot_mol_mass += ion_mass

    # Correct by -0.0025 because above volumes were measured at 25 deg C and want 20 deg C
    #  as per Svedberg, The Ultracentrifuge, 1940, Oxford Univ. Press
    vbar_prot = (numerator / prot_mol_mass) - 0.0025

    # Get convex hull area and volume 
    coords = get_coords(model_array)
    area_hull, vol_hull, Dmax = HullVolume(coords)

    # Calculate shell volume from variable of shell thickness
    # Hydration shell thickness is empirically determined to be optimal 
    # This expands each hull plane by hydration thickness
    vol_shell_wat = area_hull * 2.8

    # Calculate total volume of hydrated convex hull (including shell waters)
    vol_hyd_hull = vol_hull + vol_shell_wat

    # Estimate axial ratio 
    # Minus 3 Ang because DNA duplex a should be rod length, not diagonal
    #   of hull end vertices (Also necessary to make apoferritin axial ratio = 1)
    a = (Dmax / 2.0) - 3.0
    b = math.sqrt((3.0 * vol_hull) / (4.0 * math.pi * a))

    # But some very spherical structures may not have the diagonal a full 3 Ang longer.
    # Can't have a < b.
    if a > b:
        # Translational Shape Factor
        numerator = math.sqrt(1.0 - (b / a) ** 2)
        denominator = ((b / a) ** 0.66666667) * math.log((1 + math.sqrt(1.0 - (b / a) ** 2)) / (b / a))
        Ft = numerator / denominator
    else:
        a = b
        Ft = 1.0

    # Weight shape factor
    # Empirically found to work better with expanded volume
    #  (Many combinations tried)
    Ft = math.sqrt(Ft)

    # Axial ratio of prolate ellipsoid of same volume as convex hull
    a_b_ratio = a / b

    # Find radius of sphere of same volume as hydrated convex hull
    factor = 3.0 / (4.0 * math.pi)
    Rht = (factor * vol_hyd_hull) ** 0.333333
    # Include Shape factor to give effective hydrodynamic translational radius
    Rht = Rht * Ft
    # Rht comes as Angstrom
    # need meters in equation below
    Rh_trans = Rht * 1e-8

    # Calculate Svedberg coeff.
    eta = 0.0100194  # poise
    rho = 0.998234  # g/ml water density
    fT = 6.0 * math.pi * eta * Rh_trans
    s = prot_mol_mass * (1.0 - (vbar_prot * rho)) / (6.02214e23 * fT)

    # Calculate translational diffusion coeff.
    # R = 8.314e7
    kB = 1.381e-16
    T = 293.15
    Dt = kB * T / fT

    # Calculate fT/fo
    # Vbar is in ml/g, need A^3/Dalton
    vol_prot = prot_mol_mass * vbar_prot / 0.60224
    Ro = ((3.0 * vol_prot) / (4.0 * math.pi)) ** 0.333333
    ffo_hyd_P = Rht / Ro

    # Calculate rotational diffusion coeff.
    # Rotation more strongly affected by hydration, Halle & Davidovic, 2003
    # So increase hydration shell thickness
    # Hydration shell thickness empirically found to be optimal
    vol_shell_wat_rot = area_hull * 4.3
    vol_hyd_hull_rot = vol_hull + vol_shell_wat_rot
    Rhr = (factor * vol_hyd_hull_rot) ** 0.333333

    # Include empirical correction for shape factor to give effective 
    #  hydrodynamic rotational radius
    # This obtained by fitting to DNA duplexes
    Fr = Ft ** 4.0
    Rhr = Rhr * Fr
    # Rhr comes as Angstrom
    # need meters in equation below
    Rh_rot = Rhr * 1e-8
    fR = 8.0 * math.pi * eta * (Rh_rot ** 3.0)
    Dr = kB * T / fR
    tauC = (4.0 * math.pi * eta * (Rh_rot ** 3.0)) / (3.0 * kB * T)
    tauC = tauC * 1e9

    # Calculate intrinsic viscosity
    # From Einstein viscosity equation,
    # [n] = (2.5*6.02214e23*(4*pi*Rht^3)/3)/prot_mol_mass
    #
    # rearrange to
    # 
    # int_vis = (10.0 * math.pi * 0.602214 * ((Rht)**3.0)) / (3.0 * prot_mol_mass)
    # Deleted.
    # The above equation is valid only for spherical particles.
    # Intrinsic viscosity is very sensitive to axial ratio and whether the molecule approximates
    # a prolate or oblate ellipsoid.
    # Decided that int_vis was beyond the scope of HullRad.
    # Leaving this here for future development.
    int_vis = 0.0

    return s, Dt, Dr, vbar_prot, Rht, ffo_hyd_P, prot_mol_mass, Ro, Rhr, int_vis, a_b_ratio, \
           Ft, Rg, Dmax, tauC, asphr, AA, NA, GL, DT


def HullVolume(coords):
    """
    Uses qconvex to calculate convex hull

    coords are coords of reduced atom model.
    """
    # Call scipy's ConvexHull using coords
    # This will be an object with attributes that we want
    convex_hull = ConvexHull(coords)

    # Area of convex hull; this is a float
    area_hull = convex_hull.area

    # Volume of convex hull; this is a float
    vol_hull = convex_hull.volume

    # Find max distance between vertices for Dmax calculation
    # Necessary for length of corresponding prolate ellipsoid of revolution
    # Note that the "vertices" from ConvexHull are indices, i.e. pointers to
    # the coordinates in "coords" that serve as the vertices.
    vertices = convex_hull.vertices
    Dmax = 0.0
    for r1 in vertices:
        for r2 in vertices:
            dist = math.sqrt((coords[r1][0] - coords[r2][0]) ** 2.0 +
                             (coords[r1][1] - coords[r2][1]) ** 2.0 +
                             (coords[r1][2] - coords[r2][2]) ** 2.0)
            if dist > Dmax:
                Dmax = dist

    return area_hull, vol_hull, Dmax


def hullRad(pdb, output=True):
    r"""
    Dtrans, Drot, S and more from PDB structure using HullRad for H2O at T=20 C.

    From the abstract of [1]_ :

     ... to accurately predict the hydrodynamic properties
     of molecular structures. This method uses a convex hull model to estimate the hydrodynamic volume
     of the molecule and is orders of magnitude faster than common methods. It works well for both folded
     proteins and ensembles of conformationally heterogeneous proteins and for nucleic acids.
     Because of its simplicity and speed, the method should be useful for the modification of
     computer-generated, intrinsically disordered protein ensembles and ensembles of flexible,
     but folded, molecules in which rapid calculation of experimental parameters is needed.

    Quality of the data is comparable to HYDROPRO [2]_ but only scalar values are calculated.
    For calculation of the 6x6 diffusion tensor see :py:func:`~.bio.utilities.runHydropro`

    Parameters
    ----------
    pdb : string or IOString from MDA PDBWriter.
        Protein or DNA

    Returns
    -------
    result : dict


    Notes
    -----
    All values are calculated for H2O at T=20 degree Celsius (see [1]_)

    Coefficients keys, meaning and units (see [1]_ for detailed explanation) ::

     'Dt'         Dt                        cm^2/s;   translational diffusion coefficient
     'Rht'        Rh from Dt                Angstroms
     'Dr'         Dr                        s^-1;     rotational diffusion coefficient
     'Rhr'        Rh  from Dr               Angstroms
     'AA'         number Amino Acids
     'NA'         number Nucleotides
     'GL'         number Saccharides
     'DT'         number Detergents
     'M'          mass                      g/mol
     'vbar_prot'  v_bar specific volume     mL/g
     'Ro'         Ro(Anhydrous)             Angstroms
     'Rg'         Rg(Anhydrous)             Angstroms
     'Dmax'       Dmax                      Angstroms
     'a_b_ratio'  axial Ratio               equivalent prolate ellipsoid axial ratio
     'ffo_hyd_P'  f/fo                      Perrin friction factor
     's'          s                         sec
     'tauC'       tauC (from Rhr)           ns;  isotropic rotational correlation time (NMR)
     'asphr'      Asphericity from Gyration Tensor

    **Biological unit**
    Use a biological unit for calculation. The crystal structure is not always the biological unit
    but these can be retrieved from PDB servers as e.g.https://www.ebi.ac.uk/pdbe/ .
    See below examples and :py:func:`~.bio.mda.mergePDBModel` .



    Examples
    --------
    Based on pdb file
    ::

     import jscatter as js
     # give existing filename
     res = js.bio.hullRad('3pgk.pdb')

    For a universe from MDAnalysis. hullRad strips non protein atoms.
    ::

     import jscatter as js
     uni = js.bio.scatteringUniverse('3pgk')
     res = js.bio.hullRad(uni.select_atoms('protein'))
     # Dtrans/Drot with conversion to nm²/ps and 1/ps
     Dtrans = res['Dt'] * 1e2
     Drot = res['Dr'] * 1e-12

    Build biological unit for a tetramer using :py:func:`~.bio.mda.mergePDBModel` from downloaded biological assembly ::

     import jscatter as js

     # first get and create the biological unit ('.pdb1') of alcohol dehydrogenase (tetramer, 144 kDa)
     adh = js.bio.fetch_pdb('4w6z.pdb1')

     # the 2 dimers in are in model 1 and 2 and need to be merged into one.
     adhmerged = js.bio.mergePDBModel(adh)
     uni = js.bio.scatteringUniverse(adhmerged)
     res = js.bio.hullRad(uni.select_atoms('protein'))

     # Dtrans/Drot with conversion to nm/ps and 1/ps
     Dtrans = res['Dt'] * 1e2
     Drot = res['Dr'] * 1e-12
     Dq = js.bio.diffusionTRUnivYlm(uni,Dtrans=Dtrans,Drot=Drot)


    References
    ----------
    .. [1] HullRad: Fast Calculations of Folded and Disordered Protein and Nucleic Acid Hydrogynamic Properties
           Fleming, PJ and Fleming, KG "",
           Biophysical Journal, 2018, 114:856-869, DOI: 10.1016/j.bpj.2018.01.002

    .. [2] Calculation of hydrodynamic properties of globular proteins from their atomic-level structure
           Garcıa De La Torre, J., M. L. Huertas, and B. Carrasco.
           Biophys. J. 78:719–730, (2000)

    """
    if not isinstance(pdb, str):
        # stringIO is a bit faster and needs no file
        temp = io.StringIO()
        with MDAnalysis.coordinates.PDB.PDBWriter(temp, n_atoms=pdb.atoms.n_atoms, multiframe=False, bonds=None) as W:
            # write pdb without hydrogens
            W.write(pdb.select_atoms('not name H*'))
            temp.seek(0)
            pdblines= [t.strip() for t in temp.readlines()]

        all_atm_rec, num_MG, num_MN, model_array = model_from_pdb(pdblines)

    else:
        # Make the reduced atom model
        all_atm_rec, num_MG, num_MN, model_array = model_from_pdb(pdb)

    # Call the main function
    names = ['s', 'Dt', 'Dr', 'vbar_prot', 'Rht', 'ffo_hyd_P', 'M', 'Ro', 'Rhr', 'int_vis',
             'a_b_ratio', 'Ft', 'Rg', 'Dmax', 'tauC', 'asphr', 'AA', 'NA', 'GL', 'DT' ]
    val = Sved(all_atm_rec, num_MG, num_MN, model_array)

    # create dict for output
    res = { k: v for k, v in zip(names, val)}

    if output:
        # print coefficients
        print('  #Amino Acids    :    %9.0f' % res['AA'])
        print('  #Nucleotides    :    %9.0f' % res['NA'])
        print('  #Saccharides    :    %9.0f' % res['GL'])
        print('  #Detergents     :    %9.0f' % res['DT'])
        print('  M               :       %6.0f     g/mol' % res['M'])
        print('  v_bar           :       %6.3f     mL/g' % res['vbar_prot'])
        print('  Ro(Anhydrous)   :       %6.2f     Angstroms' % res['Ro'])
        print('  Rg(Anhydrous)   :       %6.2f     Angstroms' % res['Rg'])
        print('  Dmax            :       %6.2f     Angstroms' % res['Dmax'])
        print('  Axial Ratio     :       %6.2f' % res['a_b_ratio'])
        print('  f/fo            :       %6.2f' % res['ffo_hyd_P'])
        print('  Dt              :       %6.2e   cm^2/s' % res['Dt'])
        print('  R(Translation)  :       %6.2f     Angstroms' % res['Rht'])
        print('  s               :       %6.2e   sec' % res['s'])
        if res['a_b_ratio'] > 2.63:
            print(' ')
            print('  Caution. Axial ratio too large for accurate prediction')
            print('  of following rotational properties by HullRad.')

        print('  Dr              :       %6.2e   s^-1' % res['Dr'])
        print('  R(Rotation)     :       %6.2f     Angstroms' % res['Rhr'])
        print('  tauC            :       %6.2f     ns (from R_rotation)' % res['tauC'])
        print('  Asphericity     :       %6.2f     (from Gyration Tensor)' % res['asphr'])

    return res




# -*- coding: utf-8 -*-
#  this file is intended to be used in the debugger
# write a script that calls your function to debug it

import numpy as np
import jscatter as js

import jscatter as js
import numpy as np
uni=js.bio.scatteringUniverse('1igt')
uni.qlist=np.r_[0.025:1.5:0.1]
uni.tlist = np.r_[1:20000:20j,20000:100000:20j]

# rigid protein trans/rot diffusion
hR = js.bio.hullRad(uni.atoms)
Dtrans = hR['Dt'] * 1e2  # in nm²/ps
Drot = hR['Dr'] * 1e-12  # in 1/ps
Iqt = js.bio.intScatFuncYlm(uni.residues, Dtrans=Dtrans, Drot=Drot)

# internal domain motions
bnm = js.bio.brownianNMdiag(uni.residues, k_b=70, f_d=500)
# 1/iRT = 12 000 ps for f_d = 0.2

OU = js.bio.intScatFuncOU(bnm, [6, 7, 8], cubesize=2)


# combine diffusion and internal dynamics
IqtOU = Iqt.copy()
for i, iqtou in enumerate(IqtOU):
    ou = OU.filter(q=iqtou.q)[0]
    # combining in this case its just multiplication in time domain
    iqtou.Y = iqtou.Y * ou.Y

# show the result comparing the contributions
p=js.grace(2,0.8)
p.multi(1, 3, hgap=0)
p[0].plot(Iqt, li=-1)
p[1].plot(OU, li=-1)
p[2].plot(IqtOU, sy=[1,0.3,-1], li=-1)
for i in [0,1,2]:
    p[i].xaxis(label=r'\xt\f{} / ps', min=0, max=100000, charsize=1.5)
    p[i].yaxis(scale='log', min=0.01, max=1)
p[0].yaxis(label='I(Q,t)/I(Q,0)', charsize=1.5)
p[0].subtitle('trans/rot diffusion', size=1.5)
p[1].subtitle('internal dynamics', size=1.5)
p[2].subtitle('trans/rot diffusion + internal dynamics',size=1.5)
p[1].title('friction dominated internal dynamics in harmonic potential',size=2)
# p.save(js.examples.imagepath+'/iqtOrnsteinUhlenbeck.jpg',size=(1000/300,400/300))

















# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Physical equations and useful formulas as quadrature of vector functions, parallel execution,
viscosity, compressibility of water, scatteringLengthDensityCalc or sedimentationProfile.
Use scipy.constants for physical constants.

- Each topic is not enough for a single module, so this is a collection.
- All scipy functions can be used. See http://docs.scipy.org/doc/scipy/reference/special.html.
- Statistical functions http://docs.scipy.org/doc/scipy/reference/stats.html.

Mass and scattering length of all elements in Elements are taken from :
 - Mass: http://www.chem.qmul.ac.uk/iupac/AtWt/
 - Neutron scattering length: http://www.ncnr.nist.gov/resources/n-lengths/list.html

Units converted to amu for mass and nm for scattering length.

"""

import numpy as np

from ..libs.imagehash import imageHash
from ..parallel import *
from .formel import *
from .quadrature import *
from .functions import *
from .physics import *
from .quadrature import _getFuncCode
from ..data import Elements, vdwradii, felectron

pQFG = parQuadratureFixedGauss
pQFG = parQuadratureFixedGauss
pQFGxD = parQuadratureFixedGaussxD
pQAG = parQuadratureAdaptiveGauss
pAC = parAdaptiveCubature
pQACC = parQuadratureAdaptiveClenshawCurtis
pQS = parQuadratureSimpson
sQS = simpleQuadratureSimpson
pDA = parDistributedAverage
mPDA = multiParDistributedAverage

#: Antisymmetric Levi-Civita symbol
eijk = np.zeros((3, 3, 3))
eijk[0, 1, 2] = eijk[1, 2, 0] = eijk[2, 0, 1] = 1
eijk[0, 2, 1] = eijk[2, 1, 0] = eijk[1, 0, 2] = -1







# T

import jscatter as js


# build a complex model of different components


def particlesWithInteraction(q, Ra, Rb, molarity, bgr, contrast, collimation=None, beta=True, dw=0.15):
    """
    Ellipsoid particles with interaction including instrumental smearing as a model for e.g. proteins.

    This model is shown in example_buildComplexModel.
    The example neglects the protein exact shape and non constant scattering length density.
    Proteins are more potato shaped  and nearly never like an ellipsoid or sphere.
    So this model is only valid at low Q as an approximation.
    Nevertheless it is a general simplified particles with interaction model.

    Parameters
    ----------
    q : float
        Wavevector
    Ra,Rb : float
        Radii as in formfactor.ellipsoid in unit nm.
    molarity : float
        Concentration in mol/l
        Volume fraction is Volume*molarity.
    contrast : float
        Contrast between ellipsoid and solvent.
    bgr : float
        Background e.g. incoherent scattering
    collimation : float
        Collimation length for SANS in unit mm. For SAXS use None.
        The aperture is set to 10 mm. Detector and collimation length are the same.
    dw : float
        Width of wavelength distribution for smearing in SANS.
    beta : bool
        True include asymmetry factor beta of
        M. Kotlarchyk and S.-H. Chen, J. Chem. Phys. 79, 2461 (1983).

    Returns
    -------
        dataArray

    Notes
    -----
    Smearing is done by ::

     js.sas.resFunct(unsmeareddata, collimation, 10, collimation, 10, 0.7, dw)

    """
    # We need to multiply form factor and structure factor and add an additional background.
    # formfactor of ellipsoid returns dataArray with beta at last column.
    ff = js.ff.ellipsoid(q, Ra, Rb, SLD=contrast)
    V = ff.EllipsoidVolume
    # the structure factor returns also dataArray
    # we need to supply a radius calculated from Ra Rb, this is an assumption of effective radius for the interaction.
    R = (Ra * Rb * Rb) ** (1 / 3.)
    # the volume fraction is concentration * volume
    # the units have to be converted as V is usually nm**3 and concentration is mol/l
    sf = js.sf.PercusYevick(q, R, molarity=molarity)
    if beta:
        # beta is asymmetry factor according to M. Kotlarchyk and S.-H. Chen, J. Chem. Phys. 79, 2461 (1983).
        # correction to apply for the structure factor
        # noinspection PyProtectedMember
        sf.Y = 1 + ff._beta * (sf.Y - 1)
    #
    # molarity (mol/l) with conversion to number/nm**3 result is in cm**-1
    ff.Y = molarity * 6.023e23 / (1000 * 1e7 ** 2) * ff.Y * sf.Y + bgr
    # add parameters for later use; ellipsoid parameters are already included in ff
    # if data are saved these are included in the file as documentation
    # or can be used for further calculations e.g. if volume fraction is needed (V*molarity)
    ff.R = R
    ff.bgr = bgr
    ff.Volume = V
    ff.molarity = molarity
    # for small angle neutron scattering we may need instrument resolution smearing
    if collimation is not None:
        # For SAX we set collimation=None and this is ignored
        # For SANS we set some reasonable values as 2000 (mm) or 20000 (mm)
        # as attribute in the data we want to fit.
        result = js.sas.resFunct(ff, collimation, 10, collimation, 10, 0.7, dw)
    else:
        result = ff
    # we return the complex model for fitting
    return result



import jscatter as js
import numpy as np

# prepare profiles SANS for typical 2m and 8m measurement
# smear calls resFunc with the respective parameters; smear also works with line collimation SAXS if needed
resol8m = js.sas.prepareBeamProfile('SANS', collDist=8000., collAperture=10, detDist=8000., sampleAperture=10,
                                    wavelength=0.5, wavespread=0.2, dpixelWidth=10, dringwidth=1)


# demonstration smearing effects

# define model and use @smear to wrap it with the smearing function
# detDist and collDist will be used in js.sas.smear to change the beamprofile for smearing

@js.sas.smear(beamProfile=resol8m)
def ellipsoid(q, a, b, bgr, detDist=2000, collDist=2000.):
    elli = js.ff.ellipsoid(q, a, b)
    elli.Y = elli.Y + bgr
    return elli


# generate some smeared data, or load them from measurement
a, b = 2, 3
data = js.dL()
data.append(ellipsoid(np.r_[0.01:1:0.01], a, b, 2, detDist=8000, collDist=8000.))
data.append(ellipsoid(np.r_[0.5:5:0.05], a, b, 2, detDist=2000, collDist=2000.))

# here we compare the difference between the 2 profiles using for both the full q range
data2 = js.dL()
data2.append(ellipsoid(np.r_[0.01:5:0.02], a, b, 2, detDist=8000, collDist=8000.))
data2.append(ellipsoid(np.r_[0.01:5:0.02], a, b, 2, detDist=2000, collDist=2000.))

# plot it
p = js.grace()
ellip = js.ff.ellipsoid(np.r_[0.01:5:0.01], a, b)
ellip.Y += 2
p.plot(ellip, sy=[1, 0.3, 1], legend='unsmeared ellipsoid')
p.yaxis(label='Intensity / a.u.', scale='l', min=1, max=1e4)
p.xaxis(label=r'Q / nm\S-1', scale='n')
p.plot(data, legend='smeared $rf_detDist')
p.plot(data2[0], li=[1, 1, 4], sy=0, legend='8m smeared full range')
p.plot(data2[1], li=[3, 1, 4], sy=0, legend='2m smeared full range')
p.legend(x=2.5, y=8000)
p.title('SANS smearing of ellipsoid')
p.save('SANSsmearing.jpg')

# now we use the simulated data to fit this to a model
# the data need attributes detDist and collDist to use correct parameters in smearing
# here we add these from above in rf_detDist (set in smear)
# for a measurement this needs to be added to the experimental data
data[0].detDist = data[0].rf_detDist
data[0].collDist = data[0].rf_collDist
data[1].detDist = data[1].rf_detDist
data[1].collDist = data[1].rf_collDist


@js.sas.smear(beamProfile=resol8m)
def smearedellipsoid(q, A, a, b, bgr, detDist=2000, collDist=2000.):
    """
    The model may use all needed parameters for smearing.
    """
    ff = js.ff.ellipsoid(q, a, b)  # calc model
    ff.Y = ff.Y * A + bgr  # multiply amplitude factor and add bgr
    return ff


# fit it , here no errors
data.makeErrPlot(yscale='l', fitlinecolor=[1, 2, 5])
data.fit(smearedellipsoid, {'A': 1, 'a': 2.5, 'b': 3.5, 'bgr': 0}, {}, {'q': 'X'})
# show the unsmeared model
p = js.grace()
for oo in data:
    p.plot(oo, sy=[-1, 0.5, -1], le='lastfit smeared')
    p.plot(oo.X, oo[2], li=[3, 2, 4], sy=0, legend='lastfit unsmeared')
p.yaxis(scale='l')
p.legend()
p.save('SANSsmearingFit.jpg')



# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
.. include:: substitutions.txt

The aim of the biomacromolecules (bio) module is the calculation of static and dynamic scattering properties
of biological molecules as proteins or DNA (and later the analysis of MD simulations).

- Functions are provided to calculate **SAXS/SANS** formfactors, effective diffusion
  or intermediate scattering functions as measured by SAXS/SANS, Neutron Spinecho Spectroscopy (**NSE**),
  BackScattering (**BS**) or TimeOfFlight (**TOF**) methods.

- For the handling of atomic structures as PDB files (https://www.rcsb.org/) we use the library
  `MDAnalysis <https://www.mdanalysis.org/>`_ which allows **reading/writing of PDB files** and
  molecular dynamics trajectory files of various formats and MD simulation tools.

- PDB structures contain in most cases no hydrogen atoms.
  There are several possibilities to **add hydrogens** to existing PDB structures
  (see Notes in :py:func:`~jscatter.bio.mda.pdb2pqr`).
  We use the algorithm  pdb2pqr (allows debumping and optimization) or a simpler algorithm from PyMol.

- **Mode analysis** allows deformations of structures or calculation of dynamic properties.
  Deformation modes can be used to fit a protein structure to SAXS/SANS data (also simultaneous).

- To inspect the content of a universe or changes in a structure Pymol, VMD or other visualization viewers can be used.
  In a Jupyter notebook nglview can be used.

.. image:: ../../examples/images/mode_animation.gif
 :align: center
 :width: 50 %
 :alt: mode_animation

|mdanalysis|



"""

import numpy as np
import warnings
from .. import _platformname

try:
    # remove annoying BiopythonDeprecationWarning
    from Bio import BiopythonDeprecationWarning
    warnings.simplefilter('ignore', BiopythonDeprecationWarning)
except ModuleNotFoundError:
    pass

from .mda import *
from .mda import vdwradii_ as vdwradii
from .scatter import *
from .nma import *
from .utilities import *
from ..libs.HullRad import hullRad


# warnings.simplefilter("error", np.VisibleDeprecationWarning)
# warnings.simplefilter(action='ignore', category=np.VisibleDeprecationWarning)
# warnings.simplefilter(action='ignore', category=DeprecationWarning)





# -*- coding: utf-8 -*-
# written by Ralf Biehl at the Forschungszentrum Jülich ,
# Jülich Center for Neutron Science 1 and Institute of Complex Systems 1
#    Jscatter is a program to read, analyse and plot data
#    Copyright (C) 2015-2024  Ralf Biehl
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import inspect
import os


import numpy as np
import scipy.special as special

from .. import formel
from .. import structurefactor as sf
from ..dataarray import dataArray as dA
from .composed import multiShellCylinder, multiShellEllipsoid
from .formfactoramplitudes import fa_sphere as _fa_sphere, fq_cuboid as _fq_cuboid, fq_prism as _fq_prism
from .formfactoramplitudes import fq_triellipsoid

from .cloudscattering import cloudScattering

__all__ = ['sphere', 'disc', 'cylinder', 'cuboid', 'ellipsoid', 'triaxialEllipsoid', 'superball', 'prism']


_path_ = os.path.realpath(os.path.dirname(__file__))

# variable to allow printout for debugging as if debug:print 'message'
debug = False


def sphere(q, radius, contrast=1):
    r"""
    Scattering of a single homogeneous sphere.

    Parameters
    ----------
    q : float
        Wavevector  in units of 1/nm
    radius : float
        Radius in units nm
    contrast : float, default=1
        Difference in scattering length to the solvent = contrast

    Returns
    -------
    dataArray
        Columns [q, Iq, fa]
        Iq    scattering intensity
        - fa formfactor amplitude
        - .I0   forward scattering


    Notes
    -----
    .. math:: I(q)=  \rho^2V^2\left[\frac{3(sin(qR) - qR cos(qR))}{(qR)^3}\right]^2

    with contrast :math:`\rho` and sphere volume :math:`V=\frac{4\pi}{3}R^3`

    The first minimum of the form factor is at qR=4.493

    Examples
    --------
    ::

     import jscatter as js
     import numpy as np
     q=js.loglist(0.1,5,300)
     p=js.grace()
     R=3
     sp=js.ff.sphere(q, R)
     p.plot(sp.X*R,sp.Y,li=1)
     p.yaxis(label='I(q)',scale='l',min=1e-4,max=1e5)
     p.xaxis(label='qR',scale='l',min=0.1*R,max=5*R)
     p.legend(x=0.15,y=0.1)
     #p.save(js.examples.imagepath+'/sphere.jpg')

    .. image:: ../../examples/images/sphere.jpg
     :align: center
     :width: 50 %
     :alt: sphere


    References
    ----------
    .. [1] Guinier, A. and G. Fournet, "Small-Angle Scattering of X-Rays", John Wiley and Sons, New York, (1955).

    """
    R = radius
    qr = np.atleast_1d(q) * R
    fa0 = (4 / 3. * np.pi * R ** 3 * contrast)  # forward scattering amplitude q=0
    faQR = fa0 * _fa_sphere(qr)
    result = dA(np.c_[q, faQR ** 2, faQR].T)
    result.columnname = 'q; Iq; fa'
    result.setColumnIndex(iey=None)
    result.radius = radius
    result.I0 = fa0**2
    result.fa0 = fa0
    result.contrast = contrast
    result.modelname = inspect.currentframe().f_code.co_name
    return result


def disc(q, R, D, SLD, solventSLD=0, alpha=None):
    r"""
    Disc form factor .

    Parameters
    ----------
    q : array
        Wavevectors, units 1/nm
    R : float
        Radius in nm.
    D : float
        Thickness of the disc in units nm.
    SLD,solventSLD : float
        Scattering length density in nm^-2.
    alpha : float, [float,float] , unit rad
        Orientation, angle between the cylinder axis and the scattering vector q.
        0 means parallel, pi/2 is perpendicular
        If alpha =[start,end] is integrated between start,end
        start > 0, end < pi/2

    Notes
    -----
     For details see :py:func:`~jscatter.formfactor.composed.multiShellDisc`.

    Examples
    --------
    Simple disc.
    The short high Q modulation is caused by the radius interference.
    A radial distribution might be needed for thin discs.

    ::

     import jscatter as js
     import numpy as np
     x=np.r_[0.01:5:0.01]
     p=js.grace()

     # single disc
     bshell = js.ff.disc(x,10,3.5,6.39e-4)
     p[0].plot(bshell, le='disc')

     p[0].yaxis(label='I(q)',scale='l',min=1e-7,max=1)
     p[0].xaxis(label='q / nm\S-1',scale='l',min=0.1,max=10)
     p[0].legend(x=2,y=0.003)
     p[0].title('simple disc')
     # p.save(js.examples.imagepath+'/simpleDisc.jpg')

    .. image:: ../../examples/images/simpleDisc.jpg
     :align: center
     :width: 50 %
     :alt: simpleDisc

    """
    if alpha is None:
        alpha = [0, np.pi / 2]
    return multiShellCylinder(q, D/2, [R], [SLD], solventSLD=solventSLD, alpha=alpha)


def cylinder(q, L, radius, SLD=1e-3, solventSLD=0, alpha=None, nalpha=90, h=None):
    r"""
    Cylinder form factor including cap.

    Based on multiShellCylinder (see there for detailed description of parameters).

    Parameters
    ----------
    q : array
        Scattering vector i units 1/nm.
    L : float
        Length in nm.
    radius : float
        Radius in nm.
    h : float
        Cap geometry
    SLD : float
        Cylinder scattering length density in units nm^-2.
    solventSLD : float
        Solvent scattering length density  in units nm^-2.
    h : float, default=None
        Geometry of the cap with radii R=(r**2+h**2)**0.5 in units nm.
        h is distance of cap center with radius R from the flat cylinder cap and r as radius of the cylinder.

        - None No cap, flat end as default.
        - 0 cap radius equal cylinder radius
        - >0 cap radius larger cylinder radius as barbell
        - <0 cap radius smaller cylinder radius as lens cap
    alpha : float, [float,float] , default [0,pi/2], unit rad
        Orientation, angle between the cylinder axis and the scattering vector q in units rad.
        0 means parallel, pi/2 is perpendicular
        If alpha =[start,end] is integrated between start,end
        start > 0, end < pi/2
    nalpha : int, default 90
        Number of points in Gauss integration along alpha.

    Notes
    -----
    Compared to SASview (5.0) this yields a factor 2 less intensity.
    Correctness can be checked as the forward scattering .I0 is independent of orientation
    and should be equal V² (V is volume) if SLD=1 and solvent SLD=0.

    Definition of parameters can be seen in this figure ignoring the outer shell.
    See :py:func:`~jscatter.formfactor.composed.multiShellCylinder` :

    .. image:: barbell.png
     :align: center
     :height: 150px
     :alt: Image of barbell


    Examples
    --------
    The typical **long cylinder** formfactor with a linear region for long cylinders.
    ::

     import jscatter as js
     import numpy as np
     q=js.loglist(0.01,8,500)
     p=js.grace()
     p.multi(1,2)
     R=2
     for L in [20,40,150]:
         cc=js.ff.cylinder(q,L=L,radius=R)
         p[0].plot(cc,li=-1,sy=0,le='L ={0:.0f} R={1:.1f}'.format(L,R))
     L=60
     for R in [1,2,4]:
         cc=js.ff.cylinder(q,L=L/R**2,radius=R)
         p[1].plot(cc,li=-1,sy=0,le='L ={0:.2f} R={1:.1f}'.format(L/R**2,R))
     p[0].yaxis(label='I(q)',scale='l',min=1e-6,max=10)
     p[0].xaxis(label='q / nm\S-1',scale='l',min=0.01,max=6)
     p[1].yaxis(label='I(q)',scale='l',min=1e-7,max=1)
     p[1].xaxis(label='q / nm\S-1',scale='l',min=0.01,max=6)
     p[1].text(r'forward scattering I0\n=(SLD*L\xp\f{}R\S2\N)\S2\N = 0.035530',x=0.02,y=0.1)
     p.title('cylinder')
     p[0].legend(x=0.012,y=0.001)
     p[1].legend(x=0.012,y=0.0001)
     #p.save(js.examples.imagepath+'/cylinder.jpg')

    .. image:: ../../examples/images/cylinder.jpg
     :align: center
     :width: 50 %
     :alt: cylinder

    The following **short cylinders** highlight the peak shape which can differ from expectations.
    Dependent on R, L interferences we see flattened peaks and that consecutive peaks also differ in shape.
    ::

     import jscatter as js
     import numpy as np
     q=js.loglist(0.1,3,200)
     p=js.grace()
     L=25
     for R in np.r_[5:12:2]:
         cc=js.ff.cylinder(q,L=L,radius=R)
         p.plot(cc.X,cc.Y/cc.I0,li=[1,3,-1],sy=0,le='L ={0:.2f} R={1:.1f}'.format(L,R))
     p.yaxis(label='I(q)',scale='l')
     p.xaxis(label='q / nm\S-1',scale='n',min=0.1,max=3.3)
     p.title('short cylinders')
     p.legend(x=2,y=0.02)
     #p.save(js.examples.imagepath+'/cylindershort.jpg')

    .. image:: ../../examples/images/cylindershort.jpg
     :align: center
     :width: 50 %
     :alt: cylindershort


    References
    ----------
    .. [1] Guinier, A. and G. Fournet, "Small-Angle Scattering of X-Rays", John Wiley and Sons, New York, (1955)
    .. [2] http://www.ncnr.nist.gov/resources/sansmodels/Cylinder.html

    """
    if alpha is None:
        alpha = [0, np.pi / 2]
    return multiShellCylinder(q, L, [radius], [SLD], h=h, solventSLD=solventSLD, alpha=alpha, nalpha=nalpha)


def cuboid(q, a, b=None, c=None, SLD=1, solventSLD=0, N=30):
    r"""
    Formfactor of rectangular cuboid with different edge lengths.

    Parameters
    ----------
    q : array
        Wavevector in 1/nm
    a,b,c : float, None
        Edge length, for a=b=c its a cube, Units in nm.
        If b=None b=a.
        If c=None c=b.
    SLD : float, default =1
        Scattering length density of cuboid.unit nm^-2
        e.g. SiO2 = 4.186*1e-6 A^-2 = 4.186*1e-4 nm^-2 for neutrons
    solventSLD : float, default =0
        Scattering length density of solvent. unit nm^-2
        e.g. D2O = 6.335*1e-6 A^-2 = 6.335*1e-4 nm^-2 for neutrons
    N : int
        Order for Gaussian integration over both phi and theta.

    Returns
    -------
    dataArray
        Columns [q,Iq]
         - .I0 forward scattering
         - .edges
         - .contrast

    Notes
    -----

    .. math:: I(q)=\rho^2V_{cube}^2 \int_{0}^{2\pi}\int_{0}^{\pi} \lvert sinc(q_xa/2 )
              sinc(q_yb/2) sinc(q_zc/2)\rvert^2 \sin\theta d\theta d\phi

    with :math:`q = (q_x,q_y,q_z) = (q\sin\theta\cos\phi,q\sin\theta\sin\phi,q\cos\theta)`
    and contrast :math:`\rho` [1]_.

    In [1]_ the edge length is only half of it.

    Examples
    --------
    ::

     import jscatter as js
     import numpy as np
     q=np.r_[0.1:5:0.01]
     p=js.grace()
     p.plot(js.ff.cuboid(q,60,4,6))
     p.plot(js.ff.cuboid(q,10,4,60))
     p.plot(js.ff.cuboid(q,11,11,11),li=1)
     p.yaxis(scale='l',label='I(q)')
     p.xaxis(scale='l',label='q / nm\S-1')
     p.title('cuboid')
     #p.save(js.examples.imagepath+'/cuboid.jpg')

    .. image:: ../../examples/images/cuboid.jpg
     :width: 50 %
     :align: center
     :alt: cuboid


    References
    ----------
    .. [1] Analysis of small-angle scattering data from colloids and polymer solutions:
           modeling and least-squares fitting
           Pedersen, Jan Skov Advances in Colloid and Interface Science 70, 171 (1997)
           http://dx.doi.org/10.1016/S0001-8686(97)00312-6

    """
    if b is None:
        b = a
    if c is None:
        c = b
    sld = SLD - solventSLD
    V = a * b * c
    q = np.atleast_1d(q)

    # integrate by Gauss quadrature rule
    fq = formel.pQFGxD(_fq_cuboid, [0, 0], [np.pi/2, np.pi/2], ['p', 't'], q=q, a=a, b=b, c=c, n=N) * 8 / (4 * np.pi)

    I0 = V ** 2 * sld ** 2

    result = dA(np.c_[q, I0 * fq].T)
    result.setColumnIndex(iey=None)
    result.columnname = 'q; Iq'
    result.modelname = inspect.currentframe().f_code.co_name
    result.I0 = I0
    result.edges = [a, b, c]
    result.contrast = sld
    return result


def ellipsoid(q, Ra, Rb, SLD=1, solventSLD=0, alpha=None, tol=1e-6):
    r"""
    Form factor for a simple ellipsoid (ellipsoid of revolution).

    Parameters
    ----------
    q : float
        Scattering vector unit e.g.  1/A or 1/nm  1/Ra
    Ra : float
        Radius rotation axis   units in 1/unit(q)
    Rb : float
        Radius rotated axis    units in 1/unit(q)
    SLD : float, default =1
        Scattering length density of unit nm^-2
        e.g. SiO2 = 4.186*1e-6 A^-2 = 4.186*1e-4 nm^-2 for neutrons
    solventSLD : float, default =0
        Scattering length density of solvent. unit nm^-2
        e.g. D2O = 6.335*1e-6 A^-2 = 6.335*1e-4 nm^-2 for neutrons
    alpha : [float,float] , default [0,90]
        Angle between rotation axis Ra and scattering vector q in unit grad
        Between these angles orientation is averaged
        alpha=0 axis and q are parallel, other orientation is averaged
    tol : float
        relative tolerance for integration between alpha

    Returns
    -------
    dataArray
        Columns [q; Iq; beta ]
         - .RotationAxisRadius
         - .RotatedAxisRadius
         - .EllipsoidVolume
         - .I0         forward scattering q=0
         - beta is asymmetry factor according to [3]_.
           :math:`\beta = |<F(Q)>|^2/<|F(Q)|^2>` with scattering amplitude :math:`F(Q)` and
           form factor :math:`P(Q)=<|F(Q)|^2>`

    Notes
    -----
    See :py:func:`~jscatter.formfactor.bodies.triaxialEllipsoid` with Rb=Rc for the equation.

    Examples
    --------
    Simple ellipsoid in vacuum::

     import jscatter as js
     import numpy as np
     x=np.r_[0.1:10:0.01]
     Rp=6.
     Re=8.
     elli = js.ff.ellipsoid(x,Rp,Re,1)
     # plot it
     p=js.grace()
     p.plot(elli)
     p.yaxis(scale='l',label='I(q)',min=0.01,max=100)
     p.xaxis(scale='l',label='q / nm\S-1',min=0.1,max=10)
     p.title('ellipsoid')
     # p.save(js.examples.imagepath+'/ellipsoid.jpg')

    .. image:: ../../examples/images/ellipsoid.jpg
     :width: 50 %
     :align: center
     :alt: ellipsoid


    References
    ----------
    .. [1] Structure Analysis by Small-Angle X-Ray and Neutron Scattering
           Feigin, L. A, and D. I. Svergun, Plenum Press, New York, (1987).
    .. [2] http://www.ncnr.nist.gov/resources/sansmodels/Ellipsoid.html
    .. [3] M. Kotlarchyk and S.-H. Chen, J. Chem. Phys. 79, 2461 (1983).

    """
    if alpha is None:
        alpha = [0, 90]
    result = multiShellEllipsoid(q, Ra, Rb, shellSLD=SLD, solventSLD=solventSLD, alpha=alpha, tol=tol)
    attr = result.attr
    result.EllipsoidVolume = result.outerVolume
    result.RotationAxisRadius = Ra
    result.RotatedAxisRadius = Rb
    result.contrast = result.shellcontrast
    result.angles = alpha
    attr.remove('columnname')
    attr.remove('I0')
    for at in attr:
        delattr(result, at)
    result.modelname = inspect.currentframe().f_code.co_name
    return result


def triaxialEllipsoid(q, Ra, Rb, Rc, SLD=1, solventSLD=0, n=25):
    r"""
    Formfactor triaxial ellipsoid.

    Parameters
    ----------
    q : array
        Scattering vector in units 1/nm.
    Ra,Rb,Rc : float
        Half axes = radii in units nm.
    SLD : float
        Scattering length density in unit nm^-2.
    solventSLD : float
        Scattering length density of solvent in unit nm^-2 .
    n : int
        Order for Gaussian integration over both phi and theta.

    Returns
    -------
    dataArray
        Columns [q; Iq]
         - .Ra, .Rb, .Rc
         - .volume
         - .I0         forward scattering q=0

    Notes
    -----
    According to [1]_ the triaxial ellipsoid is

    .. math:: I(q) = V^2 \rho^2 \int_0^{2\pi} \int_0^{\pi} \phi(QR_a)^2 sin(\theta) d\theta d\phi

    with :math:`u=QR_a` and :math:`Q = (q_x^2 + (R_b/R_aq_y)^2 + (R_c/R_aq_z)^2 )^{(1/2)}`  and

    .. math:: \Phi(u) = \frac{3(sin(u)-u cos(u))}{u^3}

    also contrast :math:`\rho`, ellipsoid volume :math:`V = 4/3\pi R_a R_b R_c`


    Examples
    --------
    Triaxial ellipsoid in vacuum ::

     import jscatter as js
     import numpy as np
     q=np.r_[0.1:6:0.01]

     p=js.grace()
     elli = js.ff.triaxialEllipsoid(q,3,3,3)
     p.plot(elli, sy=3, le='sphere')
     p.plot(js.ff.sphere(q,radius=3),li=[1,2,1],sy=0)
     # rotation ellipsoid
     relli = js.ff.triaxialEllipsoid(q,3,4,4)
     p.plot(relli, sy=2, le='rot ellipsoid')
     p.plot(js.ff.ellipsoid(q,Ra=3,Rb=4),li=[1,2,1],sy=0)

     telli = js.ff.triaxialEllipsoid(q,3,4,5)
     p.plot(telli, sy=[3,0.2,3], le='triaxial ellipsoid')

     p.yaxis(scale='l',label='I(q)',min=0.01,max=1e5)
     p.xaxis(scale='l',label='q / nm\S-1',min=0.1,max=10)
     p.legend(x=0.15,y=100)
     p.title('triaxial ellipsoid model comparison ')
     p.subtitle('lines are standard models for sphere and ellipsoid')
     #p.save(js.examples.imagepath+'/triaxialEllipsoid(.jpg')

    .. image:: ../../examples/images/triaxialEllipsoid(.jpg
     :width: 50 %
     :align: center
     :alt: triaxialEllipsoid

    References
    ----------
    .. [1]  Generalizing small-angle scattering form factors with linear transformations
            Matt Thompson
            J. Appl. Cryst. (2020). 53, 1387-1391  https://doi.org/10.1107/S1600576720010389

    """
    # forward scattering Q=0 -------------
    V = 4 / 3. * np.pi * Ra * Rb * Rc
    contrast = SLD - solventSLD

    # integration over orientations for all q
    res = formel.pQFGxD(fq_triellipsoid, [0, 0], [np.pi/2, np.pi/2], ['p', 't'],  n=n, q=q, Ra=Ra, Rb=Rb, Rc=Rc)
    res *= 8 / (4 * np.pi)

    result = dA(np.c_[q, V**2 * contrast**2 * res].T)
    result.setColumnIndex(iey=None)
    result.columnname = 'q; Iq'
    result.volume = V
    result.Ra = Ra
    result.Rb = Rb
    result.Rc = Rc
    result.I0 = V**2 * contrast**2
    result.modelname = inspect.currentframe().f_code.co_name

    return result


def superball(q, R, p, SLD=1, solventSLD=0, nGrid=12, returngrid=False):
    r"""
    A superball is a general mathematical shape that can be used to describe rounded cubes, sphere and octahedron's.

    The shape parameter p continuously changes from star, octahedron, sphere to cube.

    Parameters
    ----------
    q : array
        Wavevector in 1/nm
    R : float, None
        2R = edge length
    p : float, 0<p<100
        Parameter that describes shape
         - p=0       empty space
         - p<0.5     concave octahedron's
         - p=0.5     octahedron
         - 0.5<p<1   convex octahedron's
         - p=1       spheres
         - p>1       rounded cubes
         - p->inf    cubes
    SLD : float, default =1
        Scattering length density of cuboid. unit nm^-2
    solventSLD : float, default =0
        Scattering length density of solvent. unit nm^-2
    nGrid : int
        Number of gridpoints in superball volume is ~ nGrid**3.
        The accuracy can be increased increasing the number of grid points dependent on needed q range.
        Orientational average is done with 2(nGrid*4)+1 orientations on Fibonacci lattice.
    returngrid : bool
        Return only grid as lattice object.
        The a visualisation can be done using grid.show()

    Returns
    -------
    dataArray
        Columns [q,Iq, beta]

    Notes
    -----
    The shape is described by

    .. math:: |x|^{2p} + |y|^{2p} + |z|^{2p} \le |R|^{2p}

    which results in a sphere for p=1. The numerical integration is done by a pseudorandom grid of scatterers.

    .. image:: ../../examples/images/superballfig.jpg
     :width: 100 %
     :align: center
     :alt: superballfig

    Examples
    --------
    Visualisation as shown above ::

     import jscatter as js
     import numpy as np
     import matplotlib.pyplot as plt
     from mpl_toolkits.mplot3d import Axes3D
     fig = plt.figure(figsize=[15,3])
     q=np.r_[0:5:0.1]
     R=3
     for i,p in enumerate([0.2,0.5,1,1.3,20],1):
         ax = fig.add_subplot(1,5,i, projection='3d')
         grid=js.ff.superball(q,R,p=p,nGrid=40,returngrid=True)
         grid.filter(lambda xyz: np.linalg.norm(xyz))
         grid.show(fig=fig, ax=ax,atomsize=0.2)
         ax.set_title(f'p={p:.2f}')
     #fig.savefig(js.examples.imagepath+'/superballfig.jpg')



    Compare to extreme cases of sphere (p=1) and cube (p->inf , use here 100)
    to estimate the needed accuracy in your Q range. ::

     import jscatter as js
     import numpy as np
     #
     q=np.r_[0:3.5:0.02]
     R=6
     nGrid=25
     p=js.grace()
     p.multi(2,1)
     p[0].yaxis(scale='l',label='I(q)')
     ss=js.ff.superball(q,R,p=1,nGrid=12)
     p[0].plot(ss,legend='superball p=1 nGrid=12 default')
     ss=js.ff.superball(q,R,p=1,nGrid=25)
     p[0].plot(ss,legend='superball p=1 nGrid=25')
     p[0].plot(js.ff.sphere(q,R),li=1,sy=0,legend='sphere ff')
     p[0].legend(x=2,y=5e5)
     #
     p[1].yaxis(scale='l',label='I(q)')
     p[1].xaxis(scale='n',label='q / nm\S-1')
     cc=js.ff.superball(q,R,p=100)
     p[1].plot(cc,sy=[1,0.3,1],legend='superball p=100 nGrid=12')
     cc=js.ff.superball(q,R,p=100,nGrid=25)
     p[1].plot(cc,sy=[1,0.3,2],legend='superball p=100 nGrid=25')
     p[1].plot(js.ff.cuboid(q,2*R),li=4,sy=0,legend='cuboid')
     p[1].legend(x=2,y=9e5)
     p[0].title('Superball with transition from sphere to cuboid')
     p[0].subtitle('p=1 sphere; p>1 round cube; p>20 cube  ')
     #p.save(js.examples.imagepath+'/superball.jpg')

    .. image:: ../../examples/images/superball.jpg
     :width: 50 %
     :align: center
     :alt: superball

    **Superball scaling** with :math:`q/p^{1/3}` close to sphere shape with p=1.
    Small deviations from sphere (as a kind of long wavelength roughness) cannot be discriminated from polydispersity
    or small ellipsoidality.
    ::

     import jscatter as js
     import numpy as np
     q=np.r_[0:5:0.02]
     R=3

     Fq=js.dL()
     for i,p in enumerate([0.5,0.8,0.9,1,1.115,1.3,2],1):
         fq=js.ff.superball(q,R,p=p,nGrid=20)
         Fq.append(fq)

     pp=js.grace()
     pp.multi(2,1,vgap=0.2)
     for fq in Fq[1:-1]:
         pp[0].plot(fq.X,fq.Y/fq.Y[0],sy=[-1,0.2,-1],le=f'{fq.rounding_p:.2g}')
         pp[1].plot(fq.X*fq.rounding_p**(1/3),fq.Y/fq.Y[0],sy=[-1,0.2,-1],le=f'{fq.rounding_p:.2g}')
     fq=Fq[0]
     pp[0].plot(fq.X,fq.Y/fq.Y[0],sy=0,li=[1,2,-1],le=f'{fq.rounding_p:.2g}')
     pp[1].plot(fq.X*fq.rounding_p**(1/3),fq.Y/fq.Y[0],sy=0,li=[1,2,-1],le=f'{fq.rounding_p:.2g}')
     fq=Fq[-1]
     pp[0].plot(fq.X,fq.Y/fq.Y[0],sy=0,li=[3,2,-1],le=f'{fq.rounding_p:.2g}')
     pp[1].plot(fq.X*fq.rounding_p**(1/3),fq.Y/fq.Y[0],sy=0,li=[3,2,-1],le=f'{fq.rounding_p:.2g}')

     pp[0].legend(x=0.2,y=0.05)
     pp[0].yaxis(label='I(q)',scale='l')
     pp[1].yaxis(label='I(q)',scale='l')
     pp[0].xaxis(label='q / nm')
     pp[1].xaxis(label=r'q/p\S1/3\N')
     pp[0].title('superball scaling')
     pp[0].subtitle('p=0.5 octahedron, p=1 sphere, p>10 cube')
     pp[1].text(r'q scaled by p\S-1/3\nclose to p=1 I(q) scales to similar shape ',x=4,y=0.1)
     pp[0].text('original',x=4,y=0.1)
     #p.save(js.examples.imagepath+'/superballscaling.jpg')

    .. image:: ../../examples/images/superballscaling.jpg
     :width: 50 %
     :align: center
     :alt: superballscaling



    References
    ----------
    .. [1] Periodic lattices of arbitrary nano-objects: modeling and applications for self-assembled systems
           Yager, K.G.; Zhang, Y.; Lu, F.; Gang, O.
           Journal of Applied Crystallography 2014, 47, 118–129. doi: 10.1107/S160057671302832X
    .. [2] http://gisaxs.com/index.php/Form_Factor:Superball

    """
    p2 = abs(2. * min(p, 101.))
    R = abs(R)
    q = np.atleast_1d(q)
    contrast = SLD - solventSLD
    # volume according to Soft Matter, 2012, 8, 8826-8834, DOI: 10.1039/C2SM25813G
    frac = special.gamma(1 + 1 / p2) ** 3 / special.gamma(1 + 3 / p2)
    V = 8 * R ** 3 * frac

    # superball surface radius for a point,
    # a definition of radius in p2 exponent as
    # r = lambda xyz: (np.abs(xyz[:, :3]) ** p2).sum(axis=1) ** (1. / p2)
    # The same is calculated in numpy.linalg.norm(xyz,ord=p2,axis=1) but faster

    # The integration using pseudorandom grid is as fast as 3D GaussIntegration of same quality looking at high Q
    # accuracy (deviation from analytic sphere/cube)
    # pseudorandom grid
    grid = sf.pseudoRandomLattice([2 * R, 2 * R, 2 * R], int(nGrid ** 3 / frac), b=0)
    grid.move([-R, -R, -R])  # move to zero center
    # select according to p2 norm <R
    grid.set_bsel(1, np.linalg.norm(grid.XYZall, ord=p2, axis=1) < R)
    grid.prune(grid.ball > 0)

    if returngrid:
        return grid

    # calc scattering
    result = cloudScattering(q, grid, relError=nGrid * 4)
    result.columnname = 'q; Iq; beta; fa'
    result.Y = result.Y * V ** 2 * contrast ** 2
    result.modelname = inspect.currentframe().f_code.co_name
    result.R = R
    result.Volume = V
    result.rounding_p = p2 / 2.
    result.contrast = contrast
    result.I0 = V ** 2 * contrast ** 2

    return result


def prism(q, R, H, SLD=1, solventSLD=0, relError=300):
    r"""
    Formfactor of prism (equilateral triangle) .

    Parameters
    ----------
    q : array 3xN
    R : float
        Edge length of equilateral triangle in units nm.
    H : float
        Height in units nm
    SLD : float, default =1
        Scattering length density unit nm^-2
        e.g. SiO2 = 4.186*1e-6 A^-2 = 4.186*1e-4 nm^-2 for neutrons
    solventSLD : float, default =0
        Scattering length density of solvent. unit nm^-2
        e.g. D2O = 6.335*1e-6 A^-2 = 6.335*1e-4 nm^-2 for neutrons
    relError : float, default 300
        Determines how points on sphere are selected for integration
         - >=1  Fibonacci Lattice with relError*2+1 points (min 15 points)
         - 0<1 Pseudo random points on sphere (see randomPointsOnSphere).
               Stops if relative improvement in mean is less than relError (uses steps of 20*ncpu new points).
               Final error is (stddev of N points) /sqrt(N) as for Monte Carlo methods.
               even if it is not a correct 1-sigma error in this case.

    Returns
    -------
    dataArray [q, fq]

    Notes
    -----
    With contrast :math:`\rho` and wavevector :math:`q=[q_x,q_y,q_z]` the scattering amplitude :math:`F_a(q)` is

    .. math:: F_a(q_x,q_y,q_z) = \rho \frac{2 \sqrt{3} e^{-iq_yR/ \sqrt{3}} H} {q_x (q_x^2-3q_y^2)} \
              (q_x e^{i q_yR\sqrt{3}} - q_xcos(q_xR) - i\sqrt{3} q_ysin(q_xR))  sinc(q_zH/2)

    and :math:`F(q)=<F_a(q)F^*_a(q)>=<|F_a(q)|^2>`

    Examples
    --------
    ::

     import jscatter as js
     q = js.loglist(0.1,5,100)
     p = js.grace()
     fq = js.ff.prism(q,3,3)
     p.plot(fq.X,fq.Y/fq.I0)
     p.yaxis(scale='log')

    References
    ----------
    .. [1] DNA-Nanoparticle Superlattices Formed From Anisotropic Building Blocks
          Jones et al
          Nature Materials 9, 913–917 (2010), doi: 10.1038/nmat2870

    """
    V = np.sqrt(3)*R*R*H
    sld = SLD - solventSLD
    I0 = V*V*sld*sld

    fq, err = formel.sphereAverage(funktion=_fq_prism, Q=q, R=R, H=H, passPoints=True, relError=relError).reshape(2, -1)

    result = dA(np.c_[q, sld**2 * fq].T)
    result.setColumnIndex(iey=None)
    result.columnname = 'q; Iq'
    result.modelname = inspect.currentframe().f_code.co_name
    result.I0 = I0
    result.height = H
    result.edge = R
    result.volume = V
    result.contrast = sld
    return result




import numpy as np
import matplotlib.pyplot as pyplot

pyplot.ion()

fig = pyplot.figure(figsize=(10, 5))
fig.add_subplot(1, 2, 1)
fig.add_subplot(1, 2, 2)
fig.axes[1].set_xlim((-5, 5))
fig.axes[1].set_ylim((-5, 5))

fig.axes[0].set_xlim((-5, 5))
fig.axes[0].set_ylim((-5, 5))
R = 5
x = 0
y = 0
r = 0.4
X = []
Y = []
for rr in np.r_[0:2:10j]:
    phi = np.random.rand(1) * 2 * np.pi
    x = x + r * np.sin(phi)
    y = y + r * np.cos(phi)
    X.append(x)
    Y.append(y)
    R -= r + 0.1
    print(x, y, phi)
    fig.axes[1].add_artist(pyplot.Circle((x, y), R, color='b', fill=False))
    fig.axes[1].add_artist(pyplot.Circle((x, y), 0.05, color='b', fill=False))
fig.axes[1].plot(X, Y)

R = 5
x = 0
y = 0
r = 0.4
for rr in np.r_[0:2:10j]:
    phi = np.random.rand(1) * 2 * np.pi
    x = 0
    y = 0
    R -= r + 0.1
    print(x, y, phi)
    fig.axes[0].add_artist(pyplot.Circle((x, y), R, color='b', fill=False))

# fig.savefig('MultiLamellarVesicles.png')



# Using cloudScattering as fit model.
# We have to define a model that parametrize the building of the cloud that we get a fit parameter.
# As example we use two overlapping spheres. The model can be used to fit some data.

import jscatter as js
import numpy as np

#: test if distance from point on X axis
isInside = lambda x, A, R: ((x - np.r_[A, 0, 0]) ** 2).sum(axis=1) ** 0.5 < R


#: model
def dumbbell(q, A, R1, R2, b1, b2, bgr=0, dx=0.3, relError=100):
    # A sphere distance
    # R1, R2 radii
    # b1,b2  scattering length
    # bgr background
    # dx grid distance not a fit parameter!!
    mR = max(R1, R2)
    # xyz coordinates
    grid = np.mgrid[-A / 2 - mR:A / 2 + mR:dx, -mR:mR:dx, -mR:mR:dx].reshape(3, -1).T
    insidegrid = grid[isInside(grid, -A / 2., R1) | isInside(grid, A / 2., R2)]
    # add blength column
    insidegrid = np.c_[insidegrid, insidegrid[:, 0] * 0]
    # set the corresponding blength; the order is important as here b2 overwrites b1
    insidegrid[isInside(insidegrid[:, :3], -A / 2., R1), 3] = b1
    insidegrid[isInside(insidegrid[:, :3], A / 2., R2), 3] = b2
    # and maybe a mix ; this depends on your model
    insidegrid[isInside(insidegrid[:, :3], -A / 2., R1) & isInside(insidegrid[:, :3], A / 2., R2), 3] = (b2 + b1) / 2.
    # calc the scattering
    result = js.ff.cloudScattering(q, insidegrid, relError=relError)
    result.Y += bgr
    # add attributes for later usage
    result.A = A
    result.R1 = R1
    result.R2 = R2
    result.dx = dx
    result.bgr = bgr
    result.b1 = b1
    result.b2 = b2
    result.insidegrid = insidegrid
    return result


#
# test it
q = np.r_[0.01:10:0.02]
data = dumbbell(q, 4, 2, 2, 0.5, 1.5)
#
# Fit your data like this (I know that b1 abd b2 are wrong).
# It may be a good idea to use not the highest resolution in the beginning because of speed.
# If you have a good set of starting parameters you can decrease dx.
data2 = data.prune(number=200)
data2.makeErrPlot(yscale='l')
data2.setlimit(A=[0, None, 0])
data2.fit(model=dumbbell,
          freepar={'A': 3},
          fixpar={'R1': 2, 'R2': 2, 'dx': 0.3, 'b1': 0.5, 'b2': 1.5, 'bgr': 0},
          mapNames={'q': 'X'})

# An example to demonstrate how to build a complex shaped object with a simple cubic lattice
# Methods are defined in lattice objects.
# A cube with the corners decorated by spheres

import jscatter as js
import numpy as np

grid = js.sf.scLattice(0.2, 2 * 15, b=[0])
v1 = np.r_[4, 0, 0]
v2 = np.r_[0, 4, 0]
v3 = np.r_[0, 0, 4]
grid.inParallelepiped(v1, v2, v3, b=1)
grid.inSphere(1, center=[0, 0, 0], b=2)
grid.inSphere(1, center=v1, b=3)
grid.inSphere(1, center=v2, b=4)
grid.inSphere(1, center=v3, b=5)
grid.inSphere(1, center=v1 + v2, b=6)
grid.inSphere(1, center=v2 + v3, b=7)
grid.inSphere(1, center=v3 + v1, b=8)
grid.inSphere(1, center=v3 + v2 + v1, b=9)
fig = grid.show()
js.mpl.show()



# Analyse SAS data by extrapolating the form factor followed by structure factor determination

import jscatter as js
import numpy as np

# generate some synthetic data just for this demonstration
# import the model described in example_buildComplexModel
# it describes ellipsoids with PercusYevick structure factor
from jscatter.examples import particlesWithInteraction as PWI

NN = 100
q = js.loglist(0.1, 5, NN)
data = js.dL()
bgr = js.dA(np.c_[q, 0.2 + np.random.randn(NN) * 1e-3, [1e-3] * NN].T)  # background
for m in [0.02, 0.05, 0.2, 0.6]:
    pwi = PWI(q, Ra=3, Rb=4, molarity=m * 1e-3, bgr=0, contrast=6e-4)
    pwi = pwi.addColumn(1, np.random.randn(NN) * 1e-3)
    pwi.Y = pwi.Y + bgr.Y
    pwi.setColumnIndex(iey=-1)
    data.append(pwi)

# With measured data the above is just reading data and background
# with an attribute molarity or concentration.
# This might look like this
if 0:
    data = js.dL()
    bgr = js.dA('backgroundmeasurement.dat')
    for name in ['data_conc01.dat', 'data_conc02.dat', 'data_conc05.dat', 'data_conc08.dat']:
        data.append(name)
        data[-1].molarity = float(name.split('.')[0][-2:])

p = js.grace(2, 0.8)
p.multi(1, 4)
p[0].plot(data, sy=[-1, 0.3, -1], le='c= $molarity M')
p[0].yaxis(min=1e-4, max=1e2, scale='l', label=r'I(Q) / cm\S-1', tick=[10, 9], charsize=1, ticklabel=['power', 0, 1])
p[0].xaxis(scale='l', label=r'Q / nm\S-1', min=1e-1, max=1e1, charsize=1)
p[0].text(r'original data\nlike from measurement', y=50, x=1)
p[0].legend(x=0.12, y=0.003)

# Using the synthetic data we extract again the form factor and structure factor
# subtract background and scale data by concentration or volume fraction
datas = data.copy()
for dat in datas:
    dat.Y = (dat.Y - bgr.Y) / dat.molarity
    dat.eY = (dat.eY + bgr.eY) / dat.molarity  # errors increase
p[1].plot(datas, sy=[-1, 0.3, -1], le='c= $molarity M')
p[1].yaxis(min=1, max=1e5, scale='l', tick=[10, 9], charsize=1, ticklabel=['power', 0])
p[1].xaxis(scale='l', label=r'Q / nm\S-1', min=1e-1, max=1e1, charsize=1)
p[1].text(r'bgr subtracted and\n conc. scaled', y=5e4, x=0.8)

# extrapolate to zero concentration to get the  form factor
# dataff=datas.extrapolate(molarity=0,func=lambda y:-1/y,invfunc=lambda y:-1/y)
dataff = datas.extrapolate(molarity=0)[0]
# as error *estimate* we may use the mean of the errors which is not absolutely correct
dataff = dataff.addColumn(1, datas.eY.array.mean(axis=0))
dataff.setColumnIndex(iey=2)
p[2].plot(datas[0], li=[1, 2, 4], sy=0, le='low molarity')
p[2].plot(dataff, sy=[1, 0.5, 1, 1], le='extrapolated')
p[2].yaxis(min=1, max=1e5, scale='l', tick=[10, 9], charsize=1, ticklabel=['power', 0])
p[2].xaxis(scale='l', label=r'Q / nm\S-1', min=1e-1, max=1e1, charsize=1)
p[2].legend(x=0.13, y=200)
p[2].text(r'extrapolated formfactor \ncompared to lowest conc.', y=5e4, x=0.7)

# calc the structure factor by dividing by the form factor
sf = datas.copy()
for dat in sf:
    dat.Y = dat.Y / dataff.Y
    dat.eY = (dat.eY ** 2 / dataff.Y ** 2 + dataff.eY ** 2 * (dat.Y / dataff.Y ** 2) ** 2) ** 0.5
    dat.volfrac = dat.Volume * dat.molarity
p[3].plot(sf, sy=[-1, 0.3, -1], le=r'\xF\f{}= $volfrac')
p[3].yaxis(min=0, max=2, scale='n', label=['S(Q)', 1, 'opposite'], charsize=1, ticklabel=['General', 0, 1, 'opposite'])
p[3].xaxis(scale='n', label=r'Q / nm\S-1', min=1e-1, max=2, charsize=1)
p[3].text('structure factor', y=1.5, x=1)
p[3].legend(x=0.8, y=0.5)

# remember so safe the form factor and structurefactor
# sf.save('uniquenamestructurefactor.dat')
# dataff.save('uniquenameformfactor.dat')

# save the figures
p.save('SAS_sf_extraction.agr')
p.save('SAS_sf_extraction.png', size=(2500/300, 1000/300), dpi=300)



import jscatter as js
import numpy as np

uni = js.bio.scatteringUniverse('3rn3')
uni.view()
uni.setSolvent('1d2o1')
uni.qlist = js.loglist(0.1, 5, 100)
# calc scattering both here in D2O
S = js.bio.nscatIntUniv(uni.select_atoms('protein'))
S = js.bio.xscatIntUniv(uni.select_atoms('protein'))

p = js.grace()
p.title('N scattering solvent matching')
p.yaxis(scale='l', label=r'I(Q) / nm\S2\N/particle')
p.xaxis(scale='l', label='Q / nm\S-1')

for x in np.r_[0:0.5:0.1]:
    uni.setSolvent([f'{x:.2f}d2o1',f'{1-x:.2f}h2o1' ])
    Sn = js.bio.nscatIntUniv(uni.select_atoms('protein'))
    p.plot(Sn, le=f'{x:.2f} D2O')

p.legend()
# p.save(js.examples.imagepath+'/biosolventmatching.jpg', size=(2, 2))



import jscatter as js
import numpy as np
import scipy.constants as constants

adh = js.bio.fetch_pdb('4w6z.pdb1')
# the 2 dimers in are in model 1 and 2 and need to be merged into one.
adhmerged = js.bio.mergePDBModel(adh)

p = js.grace(1.4, 1)
p.multi(1, 2)

for cc, pdb in enumerate(['3rn3', '3pgk', adhmerged], 1):
    # create universe; M and Z are for wrong named Mg and Zn  in the pdb files
    uni = js.bio.scatteringUniverse(pdb, vdwradii={'M': 1.73, 'Z': 1.7})
    uni.setSolvent('1d2o1')

    # SANS scattering intensity with conversion to 10g/l mass concentration and 1/cm units
    uni.qlist = np.r_[0.01, 0.1:3:0.03, 3:10:0.1]
    protein = uni.select_atoms('protein')
    Iq = js.bio.nscatIntUniv(protein)
    # molarity for 1g/l concentration
    mol = 1/protein.masses.sum()  # molarity 1g/l
    c = 10 * constants.N_A*mol/1000*1e-14  # conversion from 1/nm² per particle to 1/cm for 10g/l concentration
    # coherent contribution
    p[0].plot(Iq.X, Iq.Y * c, sy=0, li=[1, 3, cc], le=pdb)
    # incoherent contribution
    p[0].plot(Iq.X, Iq._P_inc * c, sy=0, li=[3, 2, cc], le='')

    # effective diffusion as seen by NSE in the initial slope of a rigid protein
    D_hr = js.bio.hullRad(uni)
    Dt = D_hr['Dt'] * 1e2  # conversion to nm²/ps
    Dr = D_hr['Dr'] * 1e-12  # conversion to 1/ps
    uni.qlist = np.r_[0.01, 0.1:6:0.06]
    Dq = js.bio.diffusionTRUnivTensor(uni.residues, DTT=Dt, DRR=Dr)
    p[1].plot(Dq.X, Dq.Y/Dq.DTTtrace, sy=[-1, 0.6, cc, cc], le=pdb)
    p[1].plot(Dq.X, Dq._Dinc_eff/Dq.DTTtrace, sy=0, li=[1, 3, cc], le=pdb)

p[0].plot([0, 10], [0.06]*2, sy=0, li=[2, 3, 1], le='')
p[0].xaxis(label=r'Q / nm\S-1', scale='log', min=0.1, max=10)
p[0].yaxis(label=r'I(Q) / 1/cm', scale='log', min=1e-4, max=3)
p[1].xaxis(label=r'Q / nm\S-1', scale='log', min=0.1, max=6)
p[1].yaxis(label=['D(Q)/D(0)', 1.5, 'opposite'], min=0.98, max=1.4)
p[0].legend(x=1, y=2)
p[0].title('neutron scattering intensity')
p[1].title('scaled effective diffusion')
p[1].subtitle('transl. diffusion at q=0, increase due to rot. diffusion')
p[1].text(r'D\sincoh', x=0.1, y=1.28, charsize=1.5)
p[1].text(r'D\scoh', x=0.3, y=1.04, charsize=1.5)
p[0].text('incoherent', x=3.7, y=0.003, charsize=1.3)
p[0].text(r'D\s2\NO background', x=1.7, y=0.038, charsize=1.3)
# p.save('bio_protein_formfactor+Diffusion.png')

